#include <gccore.h>
#include <ogcsys.h>
#include <stdarg.h>
#include "gbcore/gb.h"
#include "fontNaga10.h"
#include "screen.h"

u32 tvmode;
GXRModeObj *vmode;
u32 *xfb[2] = { NULL, NULL };
int screen_w, screen_h;

long pg_showframe;
long pg_drawframe;

void pgBitBltSgb(unsigned long x,unsigned long y,u16 *d)
{
	u32 *v0, *v1;
	u16 color;
	int r, g, b, cy, cb, cr;
	unsigned long yy;
	unsigned long xx;
	
	x/=2;
	
	for (yy=0; yy<SGBHEIGHT; yy++) {
		v0 = &xfb[pg_drawframe][(yy*2+y)*320+x];
		v1 = v0+320;
		for (xx=0; xx<SGBWIDTH; xx++){
			color = *d++;
			r = ((color      ) & 0x1F) << 3;
			g = ((color >>  5) & 0x1F) << 3;
			b = ((color >> 10) & 0x1F) << 3;
			
			cy = (299*r + 587*g + 114*b) / 1000;
			cb = (-16874*r - 33126*g + 50000*b + 12800000) / 100000;
			cr = (50000*r - 41869*g - 8131*b + 12800000) / 100000;
			
			*v0++ = *v1++ = (cy<<24) | (cb<<16) | (cy<<8) | cr;
		}
	}
}

void pgBitBltN2(unsigned long x, unsigned long y, u16 *d)
{
	u32 *v0, *v1;
	u16 color;
	int r, g, b, cy, cb, cr;
	unsigned long yy;
	unsigned long xx;
	
	x/=2;
	
	d += GUARD_LINE;
	for (yy=0; yy<GBHEIGHT; yy++) {
		v0 = &xfb[pg_drawframe][(yy*2+y)*320+x];
		v1 = v0+320;
		for (xx=0; xx<GBWIDTH; xx++){
			color = *d++;
			r = ((color      ) & 0x1F) << 3;
			g = ((color >>  5) & 0x1F) << 3;
			b = ((color >> 10) & 0x1F) << 3;
			
			cy = (299*r + 587*g + 114*b) / 1000;
			cb = (-16874*r - 33126*g + 50000*b + 12800000) / 100000;
			cr = (50000*r - 41869*g - 8131*b + 12800000) / 100000;
			
			*v0++ = *v1++ = (cy<<24) | (cb<<16) | (cy<<8) | cr;
		}
		d+=GUARD_LINE*2;
	}
}

void pgBitBltN3(unsigned long x, unsigned long y, u16 *d)
{
	u32 *v0, *v1, *v2;
	u16 color;
	int r, g, b, cy[2], cb[2], cr[2], cbb, crr;
	unsigned long yy;
	unsigned long xx;
	
	x/=2;
	
	d += GUARD_LINE;
	for (yy=0; yy<GBHEIGHT; yy++) {
		v0 = &xfb[pg_drawframe][(yy*3+y)*320+x];
		v1 = v0+320;
		v2 = v1+320;
		
		for (xx=0; xx<GBWIDTH; xx+=2){
			color = *d++;
			r = ((color      ) & 0x1F) << 3;
			g = ((color >>  5) & 0x1F) << 3;
			b = ((color >> 10) & 0x1F) << 3;
			cy[0] = (299*r + 587*g + 114*b) / 1000;
			cb[0] = (-16874*r - 33126*g + 50000*b + 12800000) / 100000;
			cr[0] = (50000*r - 41869*g - 8131*b + 12800000) / 100000;
			
			color = *d++;
			r = ((color      ) & 0x1F) << 3;
			g = ((color >>  5) & 0x1F) << 3;
			b = ((color >> 10) & 0x1F) << 3;
			cy[1] = (299*r + 587*g + 114*b) / 1000;
			cb[1] = (-16874*r - 33126*g + 50000*b + 12800000) / 100000;
			cr[1] = (50000*r - 41869*g - 8131*b + 12800000) / 100000;
			
			cbb = (cb[0]+cb[1]) >> 1;
			crr = (cr[0]+cr[1]) >> 1;
			
			*v0++ = *v1++ = *v2++ = (cy[0]<<24) | (cb[0]<<16) | (cy[0]<<8) | cr[0];
			*v0++ = *v1++ = *v2++ = (cy[0]<<24) | (cbb  <<16) | (cy[1]<<8) | crr;
			*v0++ = *v1++ = *v2++ = (cy[1]<<24) | (cb[1]<<16) | (cy[1]<<8) | cr[1];
		}
		d+=GUARD_LINE*2;
	}
}

void screenInit()
{
	VIDEO_Init();
	
	screen_w = 640;
	screen_h = 480;
	
	tvmode = VIDEO_GetCurrentTvMode();
	switch (tvmode){
	case VI_NTSC:
		vmode = &TVNtsc480IntDf;
		break;
	case VI_PAL:
		vmode = &TVPal528IntDf;
		screen_h = 528;
		break;
	case VI_MPAL:
		vmode = &TVMpal480IntDf;
		break;
	default:
		vmode = &TVNtsc480IntDf;
		break;
    }
    
	VIDEO_Configure (vmode);

	pg_showframe = 0;
	pg_drawframe = 1;

	xfb[0] = (u32 *) MEM_K0_TO_K1 (SYS_AllocateFramebuffer (vmode));
	xfb[1] = (u32 *) MEM_K0_TO_K1 (SYS_AllocateFramebuffer (vmode));
	
	console_init (xfb[0], 20, 64, vmode->fbWidth, vmode->xfbHeight, vmode->fbWidth * 2);
	
	VIDEO_ClearFrameBuffer (vmode, xfb[0], COLOR_BLACK);
	VIDEO_ClearFrameBuffer (vmode, xfb[1], COLOR_BLACK);

	VIDEO_Configure (vmode);
	VIDEO_SetNextFramebuffer (xfb[0]);
	VIDEO_SetBlack(0);
	VIDEO_Flush ();
	VIDEO_WaitVSync ();
	if (vmode->viTVMode & VI_NON_INTERLACE)
		VIDEO_WaitVSync ();
}

// by kwn
void Draw_Char_Hankaku(int x,int y,const unsigned char c,u32 col)
{
	unsigned char  *fnt;
	unsigned char  pt;
	unsigned char ch;
	int x1,y1;

	ch = c;

	// mapping
	if (ch<0x20)
		ch = 0;
	else if (ch<0x80)
		ch -= 0x20;
	else if (ch<0xa0)
		ch = 0;
	else
		ch -= 0x40;

	fnt = (unsigned char *)&hankaku_font10[ch*10];

	// draw
	u32 *dst = &xfb[pg_drawframe][y*320+x/2];
	for(y1=0;y1<10;y1++) {
		pt = *fnt++;
		for(x1=0;x1<5;x1++) {
			if (pt & 1)
				*(dst+320) = *dst = col;
			dst++;
			pt = pt >> 1;
		}
		dst = dst+640-5;
	}
}

// by kwn
void Draw_Char_Zenkaku(int x,int y,const unsigned char u,unsigned char d,u32 col)
{
	// ELISA100.FNTɑ݂Ȃ
	const unsigned short font404[] = {
		0xA2AF, 11,
		0xA2C2, 8,
		0xA2D1, 11,
		0xA2EB, 7,
		0xA2FA, 4,
		0xA3A1, 15,
		0xA3BA, 7,
		0xA3DB, 6,
		0xA3FB, 4,
		0xA4F4, 11,
		0xA5F7, 8,
		0xA6B9, 8,
		0xA6D9, 38,
		0xA7C2, 15,
		0xA7F2, 13,
		0xA8C1, 720,
		0xCFD4, 43,
		0xF4A5, 1030,
		0,0
	};
	unsigned short *fnt;
	unsigned short pt;
	int x1,y1;

	unsigned long n;
	unsigned short code;
	int i;

	// SJISR[h̐
	code = u;
	code = (code<<8) + d;

	// SJISEUCɕϊ
	if(code >= 0xE000) code-=0x4000;
	code = ((((code>>8)&0xFF)-0x81)<<9) + (code&0x00FF);
	if((code & 0x00FF) >= 0x80) code--;
	if((code & 0x00FF) >= 0x9E) code+=0x62;
	else code-=0x40;
	code += 0x2121 + 0x8080;

	// EUCbtHg̔ԍ𐶐
	n = (((code>>8)&0xFF)-0xA1)*(0xFF-0xA1)
		+ (code&0xFF)-0xA1;
	i=0;
	while(font404[i]) {
		if(code >= font404[i]) {
			if(code <= font404[i]+font404[i+1]-1) {
				n = -1;
				break;
			} else {
				n-=font404[i+1];
			}
		}
		i+=2;
	}
	fnt = (unsigned short *)&zenkaku_font10[n*10];

	// draw
	u32 *dst = &xfb[pg_drawframe][y*320+x/2];
	for(y1=0;y1<10;y1++) {
		pt = *fnt++;
		for(x1=0;x1<10;x1++) {
			if (pt & 1)
				*(dst+320) = *dst = col;
			dst++;
			pt = pt >> 1;
		}
		dst = dst+640-10;
	}
}

// by kwn
void mh_print(int x,int y,const char *msg,u32 col)
{
	int xx = x;
	unsigned char ch = 0,bef = 0, *str=(unsigned char*)msg;

	while(*str != 0) {
		ch = *str++;
		if (bef!=0) {
			Draw_Char_Zenkaku(x,y,bef,ch,col);
			x+=20;
			bef=0;
		} else {
			if (((ch>=0x80) && (ch<0xa0)) || (ch>=0xe0)) {
				bef = ch;
			} else {
				if (ch=='\n'){
					x=xx;
					y+=10;
				}else{
					Draw_Char_Hankaku(x,y,ch,col);
					x+=10;
				}
			}
		}
		if (x>=480) break;
	}
}

void pgPrintf(int x,int y,u32 color,const char *str, ...)
{
	va_list ap;
	char szBuf[512];

	va_start(ap, str);
	vsprintf(szBuf, str, ap);
	va_end(ap);

	mh_print(x,y,szBuf,color);
}
